/*
 * Copyright (c) 2010 Martin Aigner, Andreas Haas
 * http://cs.uni-salzburg.at/~maigner
 * http://cs.uni-salzburg.at/~ahaas
 *
 * University Salzburg, www.uni-salzburg.at
 * Department of Computer Science, cs.uni-salzburg.at
 */

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _SCM_DESC_H
#define	_SCM_DESC_H

#include <stdlib.h>
#include <pthread.h>
#include "stm.h"

#ifdef SCM_MT_DEBUG
#define printf printf("%lu: ", pthread_self());printf
#endif

#ifndef SCM_DESCRIPTORS_PER_PAGE
#define SCM_DESCRIPTORS_PER_PAGE \
        ((SCM_DESCRIPTOR_PAGE_SIZE - 2 * sizeof(void*))/sizeof(void*))
#endif

extern void *__real_malloc(size_t size);
extern void *__real_calloc(size_t nelem, size_t elsize);
extern void *__real_realloc(void *ptr, size_t size);
extern void __real_free(void *ptr);
extern size_t __real_malloc_usable_size(void *ptr);

/*
 * objects allocated using libscm haven an additional object header that
 * is added before the chunk returned by the malloc implemented in the
 * standard library.
 *
 * -------------------------  <- pointer to object_header_t
 * | descriptor counter dc |
 * -------------------------  <- pointer to the payload data that is
 * | payload data          |     returned to the user
 * ~ returned by malloc    ~
 * |                       |
 * -------------------------
 *
 */
#define OBJECT_HEADER(_ptr) \
    (object_header_t*)(_ptr - sizeof (object_header_t))
#define PAYLOAD_OFFSET(_o) \
    ((void*)(_o) + sizeof(object_header_t))

typedef struct descriptor_root descriptor_root_t;
typedef struct object_header object_header_t;
typedef struct descriptor_page_list descriptor_page_list_t;
typedef struct expired_descriptor_page_list expired_descriptor_page_list_t;
typedef struct descriptor_page descriptor_page_t;
typedef struct descriptor_buffer descriptor_buffer_t;

struct object_header {
    unsigned int dc;
    int finalizer_index;
};

struct descriptor_page_list {
    descriptor_page_t *first;
    descriptor_page_t *last;
};
struct expired_descriptor_page_list {
    descriptor_page_t *first; 
    descriptor_page_t *last;
    long begin;
};


/*
 * statically allocate memory for the locally clocked descriptor buffers
 * size of the locally clocked buffer is SCM_MAX_EXPIRATION_EXTENSION + 1
 * because of the additional slots for
 * 1. slot for the current time
 *
 * statically allocate memory for the globally clocked descriptor buffers
 * size of the globally clocked buffer is SCM_MAX_EXPIRATION_EXTENSION + 3
 * because of the additional slots for
 *  1. slot for the current time
 *  2. adding descriptors at current + increment + 1
 *  3. removong descriptors from current - 1
 *
 * Note: both buffers allocate SCM_MAX_EXPIRATION_EXTENSION + 3 slots for
 * page_lists but the locally clocked buffer uses only
 * SCM_MAX_EXPIRATION_EXTENSION + 1 slots
 */
struct descriptor_buffer {
    descriptor_page_list_t not_expired[SCM_MAX_EXPIRATION_EXTENSION + 3];
    unsigned int not_expired_length;
    unsigned int current_index;
};

struct descriptor_root { 
    long global_phase;

    expired_descriptor_page_list_t list_of_expired_descriptors;

    descriptor_buffer_t globally_clocked_buffer;
    descriptor_buffer_t locally_clocked_buffer;

    descriptor_page_t * descriptor_page_pool[SCM_DECRIPTOR_PAGE_FREELIST_SIZE];
    long number_of_pooled_descriptor_pages;

    //used to build a list of terminated descriptor_roots
    descriptor_root_t *next;
};

struct descriptor_page {
    descriptor_page_t *next;
    unsigned long number_of_descriptors;
    object_header_t * descriptors[SCM_DESCRIPTORS_PER_PAGE];
};

descriptor_root_t *get_descriptor_root(void)
__attribute__((visibility("hidden")));

void expire_descriptor_if_exists(expired_descriptor_page_list_t *list)
__attribute__((visibility("hidden")));

void insert_descriptor(object_header_t *o,
        descriptor_buffer_t *buffer, unsigned int expiration)
__attribute__((visibility("hidden")));

void expire_buffer(descriptor_buffer_t *buffer,
        expired_descriptor_page_list_t *exp_list)
__attribute__((visibility("hidden")));

inline void increment_current_index(descriptor_buffer_t *buffer)
__attribute__((visibility("hidden")));

int run_finalizer(object_header_t *o);

#endif	/* _SCM_DESC_H */
